"""Custom exceptions for Matchbox."""

import http
from enum import StrEnum
from typing import TYPE_CHECKING, Any

from pyarrow import Schema

if TYPE_CHECKING:
    from matchbox.common.dtos import (
        CollectionName,
        ResolutionName,
        RunID,
    )
else:
    CollectionName = Any
    ResolutionName = Any
    RunID = Any


# -- Base class for all Matchbox exceptions


class MatchboxException(Exception):
    """An exception has occurred in Matchbox.."""

    def __init__(self, message: str | None = None) -> None:
        """Initialise the exception."""
        super().__init__(message or self.__doc__)

    def to_details(self) -> dict[str, Any] | None:
        """Return exception-specific details for serialisation.

        Override in subclasses that have additional constructor arguments.
        """
        return None


class MatchboxHttpException(MatchboxException):
    """Base class for exceptions that can be transmitted over HTTP.

    Subclasses must define http_status as a class attribute.
    """

    http_status: http.HTTPStatus


# -- Common data objects exceptions --


class MatchboxNameError(MatchboxException, ValueError):
    """Name did not pass validation."""

    def __init__(self, message: str) -> None:
        """Initialise the exception."""
        super().__init__(message)


class MatchboxArrowSchemaMismatch(MatchboxException):
    """Arrow schema mismatch."""

    def __init__(self, expected: Schema, actual: Schema) -> None:
        """Initialise the exception."""
        message = f"Schema mismatch. Expected:\n{expected}\nGot:\n{actual}"

        super().__init__(message)


# -- Configuration exceptions --


class MatchboxClientSettingsException(MatchboxException):
    """Incorrect configuration provided to client."""

    def __init__(
        self,
        message: str | None = None,
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Incorrect configuration provided to client."

        super().__init__(message)


# -- Client-side API exceptions --


class MatchboxUnparsedClientRequest(MatchboxException):
    """The API could not parse the content of the client request."""

    def __init__(
        self,
        message: str | None = None,
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "The API could not parse the content of the client request"

        super().__init__(message)


class MatchboxUnhandledServerResponse(MatchboxException):
    """The API sent an unexpected response."""

    def __init__(self, http_status: int, details: str | None = None) -> None:
        """Initialise the exception."""
        message = f"The API sent an unexpected response with status {http_status}"
        if details:
            message += f" and the following message: {details}"

        super().__init__(message)


class MatchboxEmptyServerResponse(MatchboxException):
    """The server returned an empty response when data was expected."""

    def __init__(
        self, message: str | None = None, operation: str | None = None
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "The server returned an empty response when data was expected."
            if operation is not None:
                message = (
                    f"The {operation} operation returned no data from the server "
                    "when data was expected."
                )

        super().__init__(message)
        self.operation = operation


# -- SourceConfig exceptions --


class MatchboxSourceFieldError(MatchboxException):
    """Specified fields diverge with the warehouse."""


class MatchboxSourceClientError(MatchboxException):
    """Location client must be set."""


class MatchboxSourceExtractTransformError(MatchboxException):
    """Invalid ETL logic detected."""

    def __init__(
        self,
        logic: str | None = None,
    ) -> None:
        """Initialise the exception."""
        message = "Invalid ETL logic detected."
        if logic is not None:
            message += f"\n{logic}"

        super().__init__(message)


class MatchboxSourceTableError(MatchboxException):
    """Tables not found in your source data warehouse."""

    def __init__(
        self,
        message: str | None = None,
        table_name: str | None = None,
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Table doesn't exist in your source data warehouse."
            if table_name is not None:
                message += f"\nTable name: {table_name}"

        super().__init__(message)
        self.table_name = table_name


class MatchboxServerFileError(MatchboxHttpException):
    """There was a problem with file upload."""

    http_status = 400

    def __init__(self, message: str | None = None) -> None:
        """Initialise the exception."""
        if message is None:
            message = "There was a problem with file upload."

        super().__init__(message)


# -- Resource not found on server exceptions --


class MatchboxUserNotFoundError(MatchboxHttpException):
    """User not found."""

    http_status = 404

    def __init__(
        self, message: str | None = None, user_name: str | None = None
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "User not found."
            if user_name is not None:
                message = f"User {user_name} not found."

        super().__init__(message)
        self.user_name = user_name

    def to_details(self) -> dict[str, Any] | None:
        """Return user_name if set."""
        if self.user_name is not None:
            return {"user_name": self.user_name}
        return None


class MatchboxResolutionNotFoundError(MatchboxHttpException):
    """Resolution not found."""

    http_status = 404

    def __init__(
        self, message: str | None = None, name: ResolutionName | None = None
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Resolution not found."
            if name is not None:
                message = f"Resolution {name} not found."

        super().__init__(message)
        self.name = name

    def to_details(self) -> dict[str, Any] | None:
        """Return name if set."""
        if self.name is not None:
            return {"name": str(self.name)}
        return None


class MatchboxCollectionNotFoundError(MatchboxHttpException):
    """Collection not found."""

    http_status = 404

    def __init__(
        self, message: str | None = None, name: CollectionName | None = None
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Collection not found."
            if name is not None:
                message = f"Collection {name} not found."

        super().__init__(message)
        self.name = name

    def to_details(self) -> dict[str, Any] | None:
        """Return name if set."""
        if self.name is not None:
            return {"name": str(self.name)}
        return None


class MatchboxRunNotFoundError(MatchboxHttpException):
    """Run not found."""

    http_status = 404

    def __init__(self, message: str | None = None, run_id: RunID | None = None) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Run not found."
            if run_id is not None:
                message = f"Run {run_id} not found."

        super().__init__(message)
        self.run_id = run_id

    def to_details(self) -> dict[str, Any] | None:
        """Return run_id if set."""
        if self.run_id is not None:
            return {"run_id": self.run_id}
        return None


class MatchboxDataNotFound(MatchboxHttpException):
    """Data doesn't exist in the Matchbox source table."""

    http_status = 404

    def __init__(
        self,
        message: str | None = None,
        table: str | None = None,
        data: list[Any] | None = None,
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Data doesn't exist in Matchbox."
            if table is not None:
                message += f"\nTable: {table}"
            if data is not None:
                message += f"\nData: {str(data)}"

        super().__init__(message)
        self.table = table
        self.data = data

    def to_details(self) -> dict[str, Any] | None:
        """Return table and data if set."""
        details: dict[str, Any] = {}
        if self.table is not None:
            details["table"] = self.table
        if self.data is not None:
            details["data"] = self.data
        return details if details else None


# -- Server-side API exceptions --


class MatchboxConnectionError(MatchboxException):
    """Connection to Matchbox's backend database failed."""


class MatchboxPermissionDeniedError(MatchboxException):
    """Raised when a user lacks permission for an action."""


class MatchboxDeletionNotConfirmed(MatchboxHttpException):
    """Deletion must be confirmed: if certain, rerun with certain=True."""

    http_status = 409

    def __init__(
        self, message: str | None = None, children: list[str | int] | None = None
    ) -> None:
        """Initialise the exception."""
        if message is None:
            message = "Deletion must be confirmed: if certain, rerun with certain=True."

        if children is not None:
            children_names = ", ".join(str(child) for child in children)
            message = (
                f"This operation will delete the resolutions {children_names}, "
                "as well as all probabilities they have created. \n\n"
                "It won't delete validation associated with these "
                "probabilities. \n\n"
                "If you're sure you want to continue, rerun with certain=True. "
            )

        super().__init__(message)
        self.children = children

    def to_details(self) -> dict[str, Any] | None:
        """Return children if set."""
        if self.children is not None:
            return {"children": self.children}
        return None


class MatchboxResolutionAlreadyExists(MatchboxHttpException):
    """Resolution already exists."""

    http_status = 409


class MatchboxResolutionUpdateError(MatchboxHttpException):
    """Resolution metadata cannot be updated."""

    http_status = 422


class MatchboxResolutionInvalidData(MatchboxHttpException):
    """Resolution data does not match fingerprint."""

    http_status = 422


class MatchboxResolutionExistingData(MatchboxHttpException):
    """Data was already set on resolution."""

    http_status = 409


class MatchboxResolutionNotQueriable(MatchboxHttpException):
    """The resolution is not ready to be queried."""

    http_status = 422


class MatchboxCollectionAlreadyExists(MatchboxHttpException):
    """Collection already exists."""

    http_status = 409


class MatchboxRunAlreadyExists(MatchboxHttpException):
    """Run already exists."""

    http_status = 409


class MatchboxRunNotWriteable(MatchboxHttpException):
    """Run is not mutable."""

    http_status = 423


class MatchboxTooManySamplesRequested(MatchboxHttpException):
    """Too many samples have been requested from the server."""

    http_status = 422


class MatchboxGroupNotFoundError(MatchboxHttpException):
    """Raised when a group is not found."""

    http_status = 404


class MatchboxGroupAlreadyExistsError(MatchboxHttpException):
    """Raised when attempting to create a group that already exists."""

    http_status = 409


class MatchboxSystemGroupError(MatchboxException):
    """Raised when attempting to modify or delete a system group."""


# -- Adapter DB exceptions --


class MatchboxNoJudgements(MatchboxHttpException):
    """No judgements found in the database when required for operation."""

    http_status = 404


class MatchboxDatabaseWriteError(MatchboxException):
    """Could not be written to the backend DB, likely due to a constraint violation."""


class MatchboxLockError(MatchboxHttpException):
    """Trying to modify locked data."""

    http_status = 423


# -- Exception Registry --


def _get_all_subclasses(cls: type) -> list[type]:
    """Recursively get all subclasses of a class."""
    subclasses = []
    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(_get_all_subclasses(subclass))
    return subclasses


HTTP_EXCEPTION_REGISTRY: dict[str, type[MatchboxHttpException]] = {
    exc.__name__: exc for exc in _get_all_subclasses(MatchboxHttpException)
}

# Auto-generate enum from registry
_enum_members = {name: name for name in HTTP_EXCEPTION_REGISTRY}
_enum_members["MatchboxServerError"] = "MatchboxServerError"  # For generic 500 errors

MatchboxExceptionType = StrEnum("MatchboxExceptionType", _enum_members)



"""Internal Matchbox data types."""

import json
from enum import StrEnum
from typing import TYPE_CHECKING, Any, ClassVar

import polars as pl
from pydantic import BaseModel, ConfigDict, Field, model_serializer, model_validator


class _TypeNames(StrEnum):
    """Enumeration of supported data types."""

    # Boolean
    BOOLEAN = "Boolean"

    # Integers
    INT8 = "Int8"
    INT16 = "Int16"
    INT32 = "Int32"
    INT64 = "Int64"

    # Unsigned integers
    UINT8 = "UInt8"
    UINT16 = "UInt16"
    UINT32 = "UInt32"
    UINT64 = "UInt64"

    # Floating point
    FLOAT32 = "Float32"
    FLOAT64 = "Float64"

    # Decimal
    DECIMAL = "Decimal"

    # String & Binary
    STRING = "String"
    BINARY = "Binary"

    # Date & Time related
    DATE = "Date"
    TIME = "Time"
    DATETIME = "Datetime"
    DURATION = "Duration"

    # Container types
    ARRAY = "Array"
    LIST = "List"

    # Special types
    OBJECT = "Object"
    CATEGORICAL = "Categorical"
    ENUM = "Enum"
    STRUCT = "Struct"
    NULL = "Null"


class DataTypes(BaseModel):
    """Recursive definition of a data type.

    Represents Polars data types in a serialisable format.
    Can be simple (e.g., "String") or nested (e.g., List(String)).
    Arrays have a fixed shape (e.g., Array(Int64, 3) or Array(Int64, (3,))).
    Structs have named fields (e.g., Struct({"name": String, "age": Int64})).
    """

    model_config = ConfigDict(frozen=True, populate_by_name=True)

    base_type: _TypeNames = Field(alias="type")
    inner: "DataTypes | None" = None
    shape: tuple[int, ...] | None = None  # Internally stores as tuples
    fields: dict[str, "DataTypes"] | None = None  # For Struct types

    def __call__(
        self,
        inner: "DataTypes | None" = None,
        shape: tuple[int, ...] | int | None = None,
        fields: dict[str, "DataTypes"] | None = None,
    ) -> "DataTypes":
        """Allow constructing nested types.

        Examples:
            DataTypes.LIST(DataTypes.STRING)
            DataTypes.ARRAY(DataTypes.INT64, shape=3)  # Normalised to (3,)
            DataTypes.ARRAY(DataTypes.INT64, shape=(3,))
            DataTypes.STRUCT(fields={"name": DataTypes.STRING, "age": DataTypes.INT64})
        """
        # Normalise int shape to tuple for consistency with Polars
        if isinstance(shape, int):
            shape = (shape,)
        return self.model_copy(
            update={"inner": inner, "shape": shape, "fields": fields}
        )

    def _to_dict(self) -> str | dict[str, Any]:
        """Convert to dict representation for JSON serialisation.

        Returns either a string (for simple types) or a dict (for complex types).
        """
        if self.inner is None and self.fields is None:
            return self.base_type.value

        data: dict[str, Any] = {"type": self.base_type.value}

        if self.inner is not None:
            data["inner"] = self.inner._to_dict()

        if self.shape is not None:
            data["shape"] = list(self.shape)  # Convert tuple to list for JSON

        if self.fields is not None:
            data["fields"] = {
                name: dtype._to_dict() for name, dtype in self.fields.items()
            }

        return data

    @property
    def value(self) -> str:
        """Return the string representation for storage.

        - Simple types return the enum value: "String"
        - Complex types return a JSON string: '{"type": "List", "inner": "String"}'
        - Arrays include shape: '{"type": "Array", "inner": "Int64", "shape": [3]}'
        - Structs include fields: '{"type": "Struct", "fields": {"name": "String"}}'
        """
        result = self._to_dict()
        if isinstance(result, str):
            return result
        return json.dumps(result)

    def to_dtype(self) -> pl.DataType | type[pl.DataType]:
        """Convert to Polars DataType."""
        BaseClass: pl.DataType = getattr(pl, self.base_type.value, pl.String)

        # Handle Struct types with fields
        if self.base_type == _TypeNames.STRUCT and self.fields is not None:
            # Convert fields dict to list of pl.Field objects
            field_list = [
                pl.Field(name, dtype.to_dtype()) for name, dtype in self.fields.items()
            ]
            return BaseClass(field_list)

        # Handle types with inner (List, Array)
        if self.inner:
            inner_dtype = self.inner.to_dtype()
            # Array types need both inner type and shape
            if self.base_type == _TypeNames.ARRAY and self.shape is not None:
                # Polars accepts tuple directly
                return BaseClass(inner_dtype, self.shape)
            return BaseClass(inner_dtype)

        try:
            return BaseClass()
        except TypeError:
            # Some types require arguments (e.g., List)
            return BaseClass

    def to_pytype(self) -> type:
        """Convert to Python type."""
        dtype = self.to_dtype()
        if isinstance(dtype, type):
            dtype = dtype()
        return dtype.to_python()

    @classmethod
    def from_dtype(cls, dtype: pl.DataType | type[pl.DataType]) -> "DataTypes":
        """Create DataTypes from a Polars DataType."""
        base_name: str
        inner: DataTypes | None = None
        shape: tuple[int, ...] | None = None
        fields: dict[str, DataTypes] | None = None

        # Handle both Polars type classes and instances
        if isinstance(dtype, type):
            base_name = dtype.__name__
        else:
            base_name = dtype.__class__.__name__

            # Extract inner for List/Array types
            if hasattr(dtype, "inner") and dtype.inner is not None:
                inner = cls.from_dtype(dtype.inner)

            # Extract shape for Array types (Polars stores as tuple)
            if hasattr(dtype, "shape") and dtype.shape is not None:
                shape = (
                    dtype.shape if isinstance(dtype.shape, tuple) else (dtype.shape,)
                )

            # Extract fields for Struct types
            if hasattr(dtype, "fields") and dtype.fields is not None:
                fields = {
                    field.name: cls.from_dtype(field.dtype) for field in dtype.fields
                }

        # Polars uses "Utf8" internally for strings
        base_name = "String" if base_name == "Utf8" else base_name

        return cls(
            base_type=_TypeNames(base_name), inner=inner, shape=shape, fields=fields
        )

    @classmethod
    def from_pytype(cls, pytype: type) -> "DataTypes":
        """Create from Python type (current behavior)."""
        return cls.from_dtype(pl.DataType.from_python(pytype))

    @model_validator(mode="before")
    @classmethod
    def _parse_input(cls, value: Any) -> Any:  # noqa: ANN401
        """Parse string, enum, or dict inputs."""
        if isinstance(value, dict):
            # Normalise shape from int or list to tuple
            if "shape" in value:
                shape_val = value["shape"]
                if isinstance(shape_val, int):
                    value["shape"] = (shape_val,)
                elif isinstance(shape_val, list):
                    value["shape"] = tuple(shape_val)

            # Recursively parse fields if present
            if "fields" in value and isinstance(value["fields"], dict):
                value["fields"] = {
                    name: cls._parse_input(dtype)
                    if not isinstance(dtype, cls)
                    else dtype
                    for name, dtype in value["fields"].items()
                }
            return value

        if isinstance(value, _TypeNames):
            return {"type": value}

        if isinstance(value, str):
            # Try parsing JSON for nested types
            if value.startswith("{"):
                try:
                    parsed = json.loads(value)
                    return cls._parse_input(parsed)
                except json.JSONDecodeError:
                    pass
            # Simple type: "String"
            return {"type": value}

        return value

    @model_serializer
    def _serialise(self) -> str:
        """Serialise to string.

        * Simple types: "String"
        * Nested types: '{"type": "List", "inner": "String"}'
        * Arrays: '{"type": "Array", "inner": "Int64", "shape": [3]}'
        * Structs: '{"type": "Struct", "fields": {"name": "String", "age": "Int64"}}'
        """
        return self.value

    def __str__(self) -> str:
        """Human-readable string.

        * List(String)
        * Array(Int64, (3,))
        * Struct({"name": String, "age": Int64})
        """
        if self.fields:
            fields_str = (
                "{" + ", ".join(f'"{k}": {v}' for k, v in self.fields.items()) + "}"
            )
            return f"{self.base_type.value}({fields_str})"
        if self.inner:
            if self.shape is not None:
                return f"{self.base_type.value}({self.inner}, {self.shape})"
            return f"{self.base_type.value}({self.inner})"
        return self.base_type.value

    def __repr__(self) -> str:
        """Interpreter representation."""
        if not self.inner:
            return f"DataTypes.{self.base_type.name}"
        return super().__repr__()

    # Type hints for IDE autocompletion
    if TYPE_CHECKING:
        BOOLEAN: ClassVar["DataTypes"]
        INT8: ClassVar["DataTypes"]
        INT16: ClassVar["DataTypes"]
        INT32: ClassVar["DataTypes"]
        INT64: ClassVar["DataTypes"]
        UINT8: ClassVar["DataTypes"]
        UINT16: ClassVar["DataTypes"]
        UINT32: ClassVar["DataTypes"]
        UINT64: ClassVar["DataTypes"]
        FLOAT32: ClassVar["DataTypes"]
        FLOAT64: ClassVar["DataTypes"]
        DECIMAL: ClassVar["DataTypes"]
        STRING: ClassVar["DataTypes"]
        BINARY: ClassVar["DataTypes"]
        DATE: ClassVar["DataTypes"]
        TIME: ClassVar["DataTypes"]
        DATETIME: ClassVar["DataTypes"]
        DURATION: ClassVar["DataTypes"]
        ARRAY: ClassVar["DataTypes"]
        LIST: ClassVar["DataTypes"]
        OBJECT: ClassVar["DataTypes"]
        CATEGORICAL: ClassVar["DataTypes"]
        ENUM: ClassVar["DataTypes"]
        STRUCT: ClassVar["DataTypes"]
        NULL: ClassVar["DataTypes"]


# Set up enum-like class attributes: DataTypes.STRING, DataTypes.INT64, etc.
for member in _TypeNames:
    setattr(DataTypes, member.name, DataTypes(base_type=member))



"""Test the backend adapter's query functions."""

from functools import partial

import pyarrow as pa
import pyarrow.compute as pc
import pytest
from sqlalchemy import Engine
from test.fixtures.db import BACKENDS

from matchbox.common.arrow import SCHEMA_QUERY, SCHEMA_QUERY_WITH_LEAVES
from matchbox.common.dtos import Match
from matchbox.common.factories.entities import SourceEntity
from matchbox.common.factories.scenarios import setup_scenario
from matchbox.server.base import MatchboxDBAdapter


@pytest.mark.parametrize("backend", BACKENDS)
@pytest.mark.docker
class TestMatchboxQueryBackend:
    @pytest.fixture(autouse=True)
    def setup(
        self, backend_instance: MatchboxDBAdapter, sqla_sqlite_warehouse: Engine
    ) -> None:
        self.backend: MatchboxDBAdapter = backend_instance
        self.scenario = partial(setup_scenario, warehouse=sqla_sqlite_warehouse)

    def test_query_only_source(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            df_crn_sample = self.backend.query(
                source=crn_testkit.resolution_path,
                limit=10,
            )

            assert isinstance(df_crn_sample, pa.Table)
            assert df_crn_sample.num_rows == 10

            df_crn_full = self.backend.query(source=crn_testkit.resolution_path)

            assert df_crn_full.num_rows == crn_testkit.data.num_rows
            assert df_crn_full.schema.equals(SCHEMA_QUERY)

    def test_query_return_leaf_ids(self) -> None:
        """Test querying data and additionally requesting leaf IDs."""
        with self.scenario(self.backend, "index") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")

            df_crn_full = self.backend.query(
                source=crn_testkit.resolution_path, return_leaf_id=True
            )

            assert df_crn_full.num_rows == crn_testkit.data.num_rows
            assert df_crn_full.schema.equals(SCHEMA_QUERY_WITH_LEAVES)

    def test_query_with_dedupe_model(self) -> None:
        """Test querying data from a deduplication point of truth."""
        with self.scenario(self.backend, "dedupe") as dag_testkit:
            crn_testkit = dag_testkit.sources.get("crn")
            naive_crn_testkit = dag_testkit.models.get("naive_test_crn")

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=naive_crn_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            linked = dag_testkit.source_to_linked["crn"]

            assert pc.count_distinct(df_crn["id"]).as_py() == len(
                linked.true_entity_subset("crn")
            )

    def test_query_with_link_model(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            df_dh = self.backend.query(
                source=dh_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_dh, pa.Table)
            assert df_dh.num_rows == dh_testkit.data.num_rows
            assert df_dh.schema.equals(SCHEMA_QUERY)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            all_ids = pa.concat_arrays(
                [df_crn["id"].combine_chunks(), df_dh["id"].combine_chunks()]
            )

            assert pc.count_distinct(all_ids).as_py() == len(
                linked.true_entity_subset("crn", "dh")
            )

    def test_threshold_query_with_link_model(self) -> None:
        """Test querying data from a link point of truth."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "probabilistic_naive_test_crn_naive_test_cdms"
            crn_testkit = dag_testkit.sources.get("crn")
            cdms_testkit = dag_testkit.sources.get("cdms")
            linker_testkit = dag_testkit.models.get(linker_name)

            df_crn = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_crn, pa.Table)
            assert df_crn.num_rows == crn_testkit.data.num_rows
            assert df_crn.schema.equals(SCHEMA_QUERY)

            df_cdms = self.backend.query(
                source=cdms_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
            )

            assert isinstance(df_cdms, pa.Table)
            assert df_cdms.num_rows == cdms_testkit.data.num_rows
            assert df_cdms.schema.equals(SCHEMA_QUERY)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # Test query with threshold
            df_crn_threshold = self.backend.query(
                source=crn_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )
            df_cdms_threshold = self.backend.query(
                source=cdms_testkit.resolution_path,
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )
            threshold_ids = pa.concat_arrays(
                [
                    df_crn_threshold["id"].combine_chunks(),
                    df_cdms_threshold["id"].combine_chunks(),
                ]
            )

            # Query returns more clusters when threshold exceeds
            # true entity match probabilities
            assert pc.count_distinct(threshold_ids).as_py() > len(
                linked.true_entity_subset("crn", "cdms")
            )

    def test_match_one_to_many(self) -> None:
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["dh"])),
                source=dh_testkit.resolution_path,
                targets=[crn_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == dh_testkit.source.resolution_path
            assert res[0].target == crn_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["dh"]
            assert res[0].target_id == source_entity.keys["crn"]

    def test_match_many_to_one(self) -> None:
        """Test that matching data works when the source has more possible IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random many:one entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["crn"]
            assert res[0].target_id == source_entity.keys["dh"]

    def test_match_one_to_none(self) -> None:
        """Test that matching data works when the target has no IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:none entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 1},
                max_appearances={"dh": 0},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is not None
            assert res[0].source_id == source_entity.keys["crn"]
            assert res[0].target_id == source_entity.keys.get("dh", set())

    def test_match_none_to_none(self) -> None:
        """Test that matching data works when the supplied key doesn't exist."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "deterministic_naive_test_crn_naive_test_dh"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Use a non-existent source key
            non_existent_key = "foo"

            res = self.backend.match(
                key=non_existent_key,
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].cluster is None
            assert res[0].source_id == set()
            assert res[0].target_id == set()

    def test_threshold_match_many_to_one(self) -> None:
        """Test that matching data works when the target has many IDs."""
        with self.scenario(self.backend, "link") as dag_testkit:
            linker_name = "probabilistic_naive_test_crn_naive_test_cdms"
            crn_testkit = dag_testkit.sources.get("crn")
            dh_testkit = dag_testkit.sources.get("dh")
            linker_testkit = dag_testkit.models.get(linker_name)

            # Assumes CRN and DH come from same LinkedSourcesTestkit
            linked = dag_testkit.source_to_linked["crn"]

            # A random one:many entity
            source_entity: SourceEntity = linked.find_entities(
                min_appearances={"crn": 2, "dh": 1},
                max_appearances={"dh": 1},
            )[0]

            res = self.backend.match(
                key=next(iter(source_entity.keys["crn"])),
                source=crn_testkit.resolution_path,
                targets=[dh_testkit.resolution_path],
                point_of_truth=linker_testkit.resolution_path,
                threshold=100,
            )

            assert len(res) == 1
            assert isinstance(res[0], Match)
            assert res[0].source == crn_testkit.source.resolution_path
            assert res[0].target == dh_testkit.source.resolution_path
            assert res[0].source_id == source_entity.keys["crn"]
            # Match does not return true target ids when threshold
            # exceeds match probability
            assert len(res[0].target_id) < len(source_entity.keys["dh"])



"""Base class for linkers."""

import warnings
from abc import ABC, abstractmethod

import polars as pl
from pydantic import BaseModel, Field, ValidationInfo, field_validator


class LinkerSettings(BaseModel):
    """A data class to enforce basic settings dictionary shapes."""

    left_id: str = Field(
        default="id", description="The unique ID field in the left data"
    )
    right_id: str = Field(
        default="id", description="The unique ID field in the right data"
    )

    @field_validator("left_id", "right_id")
    @classmethod
    def _id_for_mb(cls, v: str, info: ValidationInfo) -> str:
        enforce = "id"
        if v != enforce:
            warnings.warn(
                f"For offline deduplication, {info.field_name} can be any field. \n\n"
                "When deduplicating to write back to the Company Matching "
                f"Framework database, the ID must be '{enforce}', generated by "
                "matchbox queries.",
                stacklevel=3,
            )
        return v


class Linker(BaseModel, ABC):
    """A base class for linkers."""

    settings: LinkerSettings

    @abstractmethod
    def prepare(self, left: pl.DataFrame, right: pl.DataFrame) -> None:
        """Prepare the linker for linking."""
        return

    @abstractmethod
    def link(self, left: pl.DataFrame, right: pl.DataFrame) -> pl.DataFrame:
        """Link the left and right dataframes."""
        return



import tempfile
from collections.abc import Generator

import pytest
from httpx import Client
from sqlalchemy import Engine

from matchbox.client.cli.eval import EntityResolutionApp
from matchbox.client.dags import DAG
from matchbox.client.eval import EvalData
from matchbox.client.locations import RelationalDBLocation
from matchbox.client.models.dedupers import NaiveDeduper
from matchbox.common.factories.sources import (
    FeatureConfig,
    SourceTestkitParameters,
    SuffixRule,
    linked_sources_factory,
)


@pytest.mark.docker
class TestE2EModelEvaluation:
    """End to end tests for model evaluation functionality."""

    def _clean_company_name(self, column: str) -> str:
        """Generate cleaning SQL for a company name column."""
        return f"""
            trim(
                regexp_replace(
                    {column},
                    ' (Ltd|Limited)$',
                    '',
                    'g'
                )
            )
        """

    @pytest.fixture(autouse=True, scope="function")
    def setup_environment(
        self,
        matchbox_client: Client,
        sqla_postgres_warehouse: Engine,
    ) -> Generator[None, None, None]:
        """Set up warehouse and database using fixtures."""
        # Persist shared setup for use in the test body
        n_true_entities = 10
        final_resolution_1_name = "final_1"
        final_resolution_2_name = "final_2"
        self.warehouse_engine = sqla_postgres_warehouse

        # Create a SINGLE source with duplicates
        source_parameters = (
            SourceTestkitParameters(
                name="source_a",
                engine=self.warehouse_engine,
                features=(
                    FeatureConfig(
                        name="company_name",
                        base_generator="company",
                    ).add_variations(
                        SuffixRule(suffix=" Ltd"),
                        SuffixRule(suffix=" Limited"),
                    ),
                    FeatureConfig(
                        name="registration_id",
                        base_generator="bothify",
                        parameters=(("text", "REG-###-???"),),
                    ),
                ),
                n_true_entities=n_true_entities,
                repetition=1,  # Duplicate all rows for deduplication
            ),
        )
        self.linked_testkit = linked_sources_factory(
            source_parameters=source_parameters, seed=42
        )
        for source_testkit in self.linked_testkit.sources.values():
            source_testkit.write_to_location()

        # Clear matchbox database before test
        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

        # Create DAG
        dw_loc = RelationalDBLocation(name="postgres").set_client(self.warehouse_engine)

        # === DAG 1: Created by User 1 (Strict Deduplication) ===
        dag1 = DAG("companies1").new_run()

        source_a_dag1 = dag1.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id, 
                    company_name, 
                    registration_id 
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id"],
        )

        source_a_dag1.query().deduper(
            name=final_resolution_1_name,
            description="Deduplicate by registration ID",
            model_class=NaiveDeduper,
            model_settings={
                "id": "id",
                "unique_fields": [source_a_dag1.f("registration_id")],
            },
        )

        dag1.run_and_sync()

        # Retain DAG for use in tests
        self.dag1 = dag1

        # === DAG 2: Created by User 2 (Loose Deduplication) ===
        dag2 = DAG("companies2").new_run()

        source_a_dag2 = dag2.source(
            location=dw_loc,
            name="source_a",
            extract_transform="""
                select
                    key::text as id, 
                    company_name, 
                    registration_id 
                from
                    source_a;
            """,
            infer_types=True,
            key_field="id",
            index_fields=["company_name", "registration_id"],
        )

        source_a_dag2.query(
            cleaning={
                "company_name": self._clean_company_name(
                    source_a_dag2.f("company_name")
                )
            }
        ).deduper(
            name=final_resolution_2_name,
            description="Deduplicate by company name",
            model_class=NaiveDeduper,
            model_settings={"id": "id", "unique_fields": ["company_name"]},
        )

        dag2.run_and_sync()

        self.dag2 = dag2

        yield

        # Teardown
        response = matchbox_client.delete("/database", params={"certain": "true"})
        assert response.status_code == 200, "Failed to clear matchbox database"

    async def in_app_evaluation(self, app: EntityResolutionApp) -> None:
        async with app.run_test() as pilot:
            await pilot.pause()

            # Verify app authenticated and loaded samples from real warehouse data
            assert app.user_name is not None
            assert len(app.queue.sessions) > 0, "Should load samples from warehouse"

            # Submit one judgement to verify data flow
            session = app.queue.sessions[0]
            for i in range(len(session.item.get_unique_record_groups())):
                session.assignments[i] = "a"  # Assign all to same cluster

            initial_judgements = EvalData().judgements
            initial_count = len(initial_judgements)

            await app.action_submit()

            final_judgements = EvalData().judgements
            assert len(final_judgements) == initial_count + 1, (
                "Judgement should flow through to backend"
            )

    @pytest.mark.asyncio
    async def test_evaluation_workflow_server(self) -> None:
        """Test end-to-end data pipeline: DAG → samples → judgement → model comparison.

        Samples clusters from the server.

        This test focuses on the full data flow through the system with real warehouse
        data, multiple DAGs, and model comparison. UI interaction details are tested
        separately in unit/integration tests.
        """
        # Load DAG from server with warehouse location
        dag: DAG = (
            DAG(str(self.dag1.name)).load_pending().set_client(self.warehouse_engine)
        )

        # Create app and verify it can load samples from real data
        app = EntityResolutionApp(
            resolution=dag.final_step.resolution_path,
            num_samples=2,
            session_tag="eval_session1",
            user="alice",
            dag=dag,
        )

        await self.in_app_evaluation(app)

        # Can filter judgements by tag
        assert len(EvalData("eval_session1").judgements)
        assert not len(EvalData("mispelled").judgements)

    @pytest.mark.asyncio
    async def test_evaluation_workflow_local(self) -> None:
        """Test end-to-end data pipeline: DAG → samples → judgement → model comparison.

        Generates a local sample file.

        This test focuses on the full data flow through the system with real warehouse
        data, multiple DAGs, and model comparison. UI interaction details are tested
        separately in unit/integration tests.
        """
        # Load DAG from server with warehouse location
        dag: DAG = (
            DAG(str(self.dag1.name)).load_pending().set_client(self.warehouse_engine)
        )
        rm = dag.resolve()

        with tempfile.NamedTemporaryFile(suffix=".pq") as tmp_file:
            # Write the parquet data to the temporary file
            rm.as_dump().write_parquet(tmp_file.name)

            # Create app and verify it can load samples
            app = EntityResolutionApp(
                num_samples=2,
                session_tag="eval_session1",
                user="alice",
                dag=dag,
                sample_file=tmp_file.name,
            )

            await self.in_app_evaluation(app)

        # Can filter judgements by tag
        assert len(EvalData("eval_session1").judgements)
        assert not len(EvalData("mispelled").judgements)



import zipfile
from io import BytesIO
from unittest.mock import Mock

import pyarrow as pa
import pyarrow.parquet as pq
import pytest
from fastapi.testclient import TestClient

from matchbox.common.arrow import (
    SCHEMA_CLUSTER_EXPANSION,
    SCHEMA_EVAL_SAMPLES,
    SCHEMA_JUDGEMENTS,
    JudgementsZipFilenames,
)
from matchbox.common.eval import Judgement
from matchbox.common.exceptions import (
    MatchboxDataNotFound,
    MatchboxResolutionNotFoundError,
    MatchboxTooManySamplesRequested,
    MatchboxUserNotFoundError,
)


def test_insert_judgement_ok(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test that a judgement is passed on to backend."""
    test_client, mock_backend, _ = api_client_and_mocks
    judgement = Judgement(user_name="alice", shown=[1, 2], endorsed=[[1], [2]])
    response = test_client.post("/eval/judgements", json=judgement.model_dump())
    assert response.status_code == 201
    assert (
        mock_backend.insert_judgement.call_args_list[0].kwargs["judgement"] == judgement
    )


def test_insert_judgement_error(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test that judgement insertion bubbles up errors."""
    test_client, mock_backend, _ = api_client_and_mocks
    fake_judgement = Judgement(
        user_name="alice", shown=[1, 2], endorsed=[[1], [2]]
    ).model_dump()

    mock_backend.insert_judgement = Mock(side_effect=MatchboxDataNotFound)
    response = test_client.post("/eval/judgements", json=fake_judgement)
    assert response.status_code == 404
    assert response.json()["exception_type"] == "MatchboxDataNotFound"

    mock_backend.insert_judgement = Mock(side_effect=MatchboxUserNotFoundError)
    response = test_client.post("/eval/judgements", json=fake_judgement)
    assert response.status_code == 404
    assert response.json()["exception_type"] == "MatchboxUserNotFoundError"


def test_get_judgements(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test that all judgements can be retrieved."""
    judgements = pa.Table.from_pylist(
        [
            {"user_name": "alice", "shown": 10, "endorsed": 11},
            {"user_name": "alice", "shown": 10, "endorsed": 12},
        ],
        schema=SCHEMA_JUDGEMENTS,
    )
    # There will be nulls in case of a schema mismatch
    assert len(judgements.drop_null()) == len(judgements)

    expansion = pa.Table.from_pylist(
        [
            {"root": 10, "leaves": [1, 2, 3]},
            {"root": 11, "leaves": [1]},
            {"root": 12, "leaves": [2, 3]},
        ],
        schema=SCHEMA_CLUSTER_EXPANSION,
    )
    # There will be nulls in case of a schema mismatch
    assert len(expansion.drop_null()) == len(expansion)

    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_judgements = Mock(return_value=(judgements, expansion))

    # Process response
    response = test_client.get("/eval/judgements")
    zip_bytes = BytesIO(response.content)

    with zipfile.ZipFile(zip_bytes, "r") as zip_file:
        with zip_file.open(JudgementsZipFilenames.JUDGEMENTS) as f1:
            downloaded_judgements = pq.read_table(f1)

        with zip_file.open(JudgementsZipFilenames.EXPANSION) as f2:
            downloaded_expansion = pq.read_table(f2)

    # Check response
    assert response.status_code == 200
    assert downloaded_judgements.equals(judgements)
    assert downloaded_expansion.equals(expansion)


def test_get_samples(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test that samples can be requested."""
    sample = pa.Table.from_pylist(
        [
            {"root": 10, "leaf": 1, "key": "1", "source": "source_a"},
            {"root": 10, "leaf": 1, "key": "2", "source": "source_a"},
        ],
        schema=SCHEMA_EVAL_SAMPLES,
    )
    # There will be nulls in case of a schema mismatch
    assert len(sample.drop_null()) == len(sample)

    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.sample_for_eval = Mock(return_value=sample)

    # Process response
    response = test_client.get(
        "/eval/samples",
        params={
            "collection": "test_collection",
            "run_id": 1,
            "resolution": "a",
            "n": 10,
            "user_name": "alice",
        },
    )
    assert response.status_code == 200

    buffer = BytesIO(response.content)
    table = pq.read_table(buffer)
    assert table.equals(sample)


@pytest.mark.parametrize(
    ("exception", "expected_exception_name"),
    [
        pytest.param(
            MatchboxUserNotFoundError,
            "MatchboxUserNotFoundError",
            id="user_not_found",
        ),
        pytest.param(
            MatchboxResolutionNotFoundError,
            "MatchboxResolutionNotFoundError",
            id="resolution_not_found",
        ),
    ],
)
def test_get_samples_404(
    exception: type[Exception],
    expected_exception_name: str,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test errors in requesting samples."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.sample_for_eval = Mock(side_effect=exception)

    response = test_client.get(
        "/eval/samples",
        params={
            "collection": "test_collection",
            "run_id": 1,
            "resolution": "a",
            "n": 10,
            "user_name": "alice",
        },
    )

    assert response.status_code == 404
    assert response.json()["exception_type"] == expected_exception_name


def test_get_samples_422(api_client_and_mocks: tuple[TestClient, Mock, Mock]) -> None:
    """Test errors in requesting samples."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.sample_for_eval = Mock(side_effect=MatchboxTooManySamplesRequested)

    response = test_client.get(
        "/eval/samples",
        params={
            "collection": "test_collection",
            "run_id": 1,
            "resolution": "a",
            "n": 10,
            "user_name": "alice",
        },
    )

    assert response.status_code == 422
    assert response.json()["exception_type"] == "MatchboxTooManySamplesRequested"



"""Minimal widget tests - verify custom configuration and display logic."""

from unittest.mock import Mock

from matchbox.client.cli.eval.widgets.assignment import AssignmentBar
from matchbox.client.cli.eval.widgets.table import ComparisonDisplayTable


class TestComparisonDisplayTable:
    """Test table widget configuration and display logic."""

    def test_build_header_formatting(self) -> None:
        """Test that headers correctly format numbers, counts, and symbols."""
        table = ComparisonDisplayTable()

        # 1. Basic Numbering
        # Position 1, single record -> "1"
        header = table._build_header(1, [1], None)
        assert header.plain == "1"

        # 2. Duplicate Counts
        # Position 2, 3 duplicate records -> "2 (×3)"
        header = table._build_header(2, [1, 2, 3], None)
        assert header.plain == "2 (×3)"

        # 3. Assignments & Symbols
        # Group 'a' (◇ diamond), Position 1 -> "◇ 1"
        header = table._build_header(1, [1], "a")
        assert "◇" in header.plain
        assert "1" in header.plain

        # Group 'b' (⬤ circle), Position 2, duplicates -> "⬤ 2 (×2)"
        header = table._build_header(2, [1, 2], "b")
        assert "⬤" in header.plain
        assert "2 (×2)" in header.plain

    def test_build_header_dimming(self) -> None:
        """Test that unassigned headers are dimmed and assigned ones are colored."""
        table = ComparisonDisplayTable()

        # Unassigned: Should be dim
        header_unassigned = table._build_header(1, [1], None)
        # Rich text spans structure: Span(start, end, style)
        # We expect 'dim' style on the number
        assert "dim" in str(header_unassigned.spans)

        # Assigned: Should NOT be dim (will have group color)
        header_assigned = table._build_header(1, [1], "a")
        # Should not have dim style (it gets stylized with group color)
        assert "dim" not in str(header_assigned.spans)


class TestAssignmentBar:
    """Test assignment bar widget status display."""

    def test_initializes_empty(self) -> None:
        """Test bar starts empty."""
        bar = AssignmentBar()
        assert bar.positions == []

    def test_reset(self) -> None:
        """Test reset creates correct number of empty slots."""
        bar = AssignmentBar()
        bar.update = Mock()  # Mock update to prevent rendering errors/side effects

        bar.reset(5)
        assert len(bar.positions) == 5
        assert all(p is None for p in bar.positions)

    def test_set_position_updates_state(self) -> None:
        """Test setting a position updates internal state."""
        bar = AssignmentBar()
        bar.update = Mock()
        bar.reset(3)

        bar.set_position(1, "a", "red")

        assert bar.positions[0] is None
        assert bar.positions[1] is not None
        assert bar.positions[1].letter == "a"
        assert bar.positions[1].colour == "red"

    def test_render_bar_logic(self) -> None:
        """Test the visual rendering logic (letters vs dots)."""
        bar = AssignmentBar()
        bar.update = Mock()

        # 1. Initial empty state
        bar.reset(3)
        bar.update.assert_called_with("[dim]•[/dim][dim]•[/dim][dim]•[/dim]")

        # 2. Set middle to 'a' -> shows letter 'a'
        bar.set_position(1, "a", "red")
        bar.update.assert_called_with("[dim]•[/dim][red]a[/red][dim]•[/dim]")

        # 3. Set last to 'a' -> adjacent same group shows dot '•'
        bar.set_position(2, "a", "red")
        bar.update.assert_called_with("[dim]•[/dim][red]a[/red][red]•[/red]")

        # 4. Set first to 'b' -> new group shows letter 'b'
        bar.set_position(0, "b", "blue")
        bar.update.assert_called_with("[blue]b[/blue][red]a[/red][red]•[/red]")



"""Utilities for inserting data into the PostgreSQL backend."""

from collections.abc import Iterator

import polars as pl
import pyarrow as pa
from sqlalchemy import exists, func, join, literal, select, update
from sqlalchemy.dialects.postgresql import (
    ARRAY,
    BIGINT,
    BOOLEAN,
    BYTEA,
    SMALLINT,
    TEXT,
    insert,
)

from matchbox.common.dtos import (
    ModelResolutionPath,
    ResolutionType,
    SourceResolutionPath,
)
from matchbox.common.exceptions import (
    MatchboxResolutionExistingData,
    MatchboxResolutionInvalidData,
)
from matchbox.common.hash import IntMap, hash_arrow_table, hash_model_results
from matchbox.common.logging import logger
from matchbox.common.transform import Cluster, DisjointSet
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import (
    Clusters,
    ClusterSourceKey,
    Contains,
    Probabilities,
    Resolutions,
    Results,
    SourceConfigs,
)
from matchbox.server.postgresql.utils.db import (
    ingest_to_temporary_table,
)
from matchbox.server.postgresql.utils.query import get_parent_clusters_and_leaves

MODEL_CLUSTERS_SCHEMA = {
    "cluster_id": pl.Int64,
    "cluster_hash": pl.Binary,
    "leaves": pl.List(pl.Int64),
    "probability": pl.Int8,
}


def insert_hashes(
    path: SourceResolutionPath, data_hashes: pa.Table, batch_size: int
) -> None:
    """Indexes hash data for a source.

    Args:
        path: The path of the source resolution
        data_hashes: Arrow table containing hash data
        batch_size: Batch size for bulk operations

    Raises:
        MatchboxResolutionNotFoundError: If the specified resolution doesn't exist.
        MatchboxResolutionInvalidData: If data fingerprint conflicts with resolution.
        MatchboxResolutionExistingData: If data was already inserted for resolution.
    """
    log_prefix = f"Index hashes {path}"
    if data_hashes.num_rows == 0:
        logger.info("No hashes given.", prefix=log_prefix)
        return

    fingerprint = hash_arrow_table(data_hashes)

    with MBDB.get_session() as session:
        resolution = Resolutions.from_path(
            path=path, res_type=ResolutionType.SOURCE, session=session
        )
        # Check if the content hash is the same
        if resolution.fingerprint != fingerprint:
            raise MatchboxResolutionInvalidData

        # Determine if the resolution already has any keys
        existing_keys = session.execute(
            select(func.count())
            .select_from(
                join(
                    ClusterSourceKey,
                    SourceConfigs,
                    ClusterSourceKey.source_config_id == SourceConfigs.source_config_id,
                )
            )
            .where(SourceConfigs.resolution_id == resolution.resolution_id)
        ).scalar_one()

        if existing_keys > 0:
            raise MatchboxResolutionExistingData

        source_config_id = resolution.source_config.source_config_id

    with (
        ingest_to_temporary_table(
            table_name="incoming_hashes",
            schema_name="mb",
            column_types={"hash": BYTEA(), "keys": ARRAY(TEXT)},
            data=data_hashes.select(["hash", "keys"]),
            max_chunksize=batch_size,
        ) as incoming,
        MBDB.get_session() as session,
    ):
        try:
            # Add clusters
            new_hashes = (
                select(incoming.c.hash)
                .distinct()
                .where(
                    ~exists(select(1).where(Clusters.cluster_hash == incoming.c.hash))
                )
            )

            stmt_insert_clusters = (
                insert(Clusters)
                .from_select(["cluster_hash"], new_hashes)
                .on_conflict_do_nothing(index_elements=[Clusters.cluster_hash])
            )

            result = session.execute(stmt_insert_clusters)
            logger.info(
                f"Will add {result.rowcount:,} entries to Clusters table",
                prefix=log_prefix,
            )
            session.flush()

            # Add source keys
            exploded = select(
                Clusters.cluster_id,
                literal(source_config_id, BIGINT).label("source_config_id"),
                func.unnest(incoming.c["keys"]).label("key"),
            ).select_from(
                incoming.join(Clusters, Clusters.cluster_hash == incoming.c.hash)
            )

            stmt_insert_keys = insert(ClusterSourceKey).from_select(
                ["cluster_id", "source_config_id", "key"],
                exploded,
            )

            result = session.execute(stmt_insert_keys)
            logger.info(
                f"Will add {result.rowcount:,} entries to ClusterSourceKey table",
                prefix=log_prefix,
            )
            session.commit()

        except Exception as e:
            # Log the error and rollback
            logger.warning(f"Error, rolling back: {e}", prefix=log_prefix)
            session.rollback()
            raise

    MBDB.vacuum_analyze(
        Clusters.__table__.fullname,
        ClusterSourceKey.__table__.fullname,
    )

    logger.info("Finished", prefix=log_prefix)


def _build_cluster_objects(
    nested_dict: dict[int, dict[str, list[dict]]],
    intmap: IntMap,
) -> dict[int, Cluster]:
    """Convert the nested dictionary to Cluster objects.

    Args:
        nested_dict: Dictionary from get_parent_clusters_and_leaves()
        intmap: IntMap object for creating new IDs safely

    Returns:
        Dict mapping cluster IDs to Cluster objects
    """
    cluster_lookup: dict[int, Cluster] = {}

    for cluster_id, data in nested_dict.items():
        # Create leaf clusters on-demand
        leaves = []
        for leaf_data in data["leaves"]:
            leaf_id = leaf_data["leaf_id"]
            if leaf_id not in cluster_lookup:
                cluster_lookup[leaf_id] = Cluster(
                    id=leaf_id, hash=leaf_data["leaf_hash"], intmap=intmap
                )
            leaves.append(cluster_lookup[leaf_id])

        # Create parent cluster
        cluster_lookup[cluster_id] = Cluster(
            id=cluster_id,
            hash=data["root_hash"],
            probability=data["probability"],
            leaves=leaves,
            intmap=intmap,
        )

    return cluster_lookup


def _results_to_cluster_pairs(
    cluster_lookup: dict[int, Cluster],
    results: pa.Table,
) -> Iterator[tuple[Cluster, Cluster, int]]:
    """Convert the results from a PyArrow table to an iterator of cluster pairs.

    Args:
        cluster_lookup (dict[int, Cluster]): A dictionary mapping cluster IDs to
            Cluster objects.
        results (pa.Table): The PyArrow table containing the results: left_id
            right_id, and probability.

    Returns:
        list[tuple[Cluster, Cluster, int]]: An iterator of tuples, each containing
            the left cluster, right cluster, and the probability, in descending
            order of probability.
    """
    for row in pl.from_arrow(results).sort("probability", descending=True).iter_rows():
        left_cluster: Cluster = cluster_lookup[row[0]]
        right_cluster: Cluster = cluster_lookup[row[1]]

        yield left_cluster, right_cluster, row[2]


def _build_cluster_hierarchy(
    cluster_lookup: dict[int, Cluster], probabilities: pa.Table
) -> pl.DataFrame:
    """Build cluster hierarchy using disjoint sets and probability thresholding.

    Args:
        cluster_lookup: Dictionary mapping cluster IDs to Cluster objects
        probabilities: Arrow table containing probability data

    Returns:
        DataFrame with MODEL_CLUSTERS_SCHEMA
    """
    logger.debug("Computing hierarchies")

    djs = DisjointSet[Cluster]()
    all_clusters: dict[bytes, Cluster] = {}
    seen_components: set[frozenset[Cluster]] = set()
    threshold: int = int(pa.compute.max(probabilities["probability"]).as_py())

    def _process_components(probability: int) -> None:
        """Process components at the current threshold."""
        components: set[frozenset[Cluster]] = {
            frozenset(component) for component in djs.get_components()
        }
        for component in components.difference(seen_components):
            cluster = Cluster.combine(
                clusters=component,
                probability=probability,
            )
            all_clusters[cluster.hash] = cluster

        return components

    for left_cluster, right_cluster, probability in _results_to_cluster_pairs(
        cluster_lookup, probabilities
    ):
        if probability < threshold:
            # Process the components at the previous threshold
            seen_components.update(_process_components(threshold))
            threshold = probability

        djs.union(left_cluster, right_cluster)

    # Process any remaining components
    _process_components(probability)
    return pl.DataFrame(
        [
            {
                "cluster_hash": cluster.hash,
                "cluster_id": cluster.id,
                "probability": cluster.probability,
                "leaves": [leaf.id for leaf in cluster.leaves]
                if cluster.leaves
                else [],
            }
            for cluster in all_clusters.values()
        ],
        schema=MODEL_CLUSTERS_SCHEMA,
    )


def insert_results(
    path: ModelResolutionPath,
    results: pa.Table,
    batch_size: int,
) -> None:
    """Writes a results table to Matchbox.

    The PostgreSQL backend stores clusters in a hierarchical structure, where
    each component references its parent component at a higher threshold.

    This means two-item components are synonymous with their original pairwise
    probabilities.

    This allows easy querying of clusters at any threshold.

    Args:
        path: The path of the model resolution to upload results for
        results: A PyArrow results table with left_id, right_id, probability
        batch_size: Number of records to insert in each batch

    Raises:
        MatchboxResolutionNotFoundError: If the specified resolution doesn't exist.
        MatchboxResolutionInvalidData: If data fingerprint conflicts with resolution.
        MatchboxResolutionExistingData: If data was already inserted for resolution.
    """
    log_prefix = f"Model {path.name}"
    if results.num_rows == 0:
        logger.info("Empty results given.", prefix=log_prefix)
        return

    resolution = Resolutions.from_path(path=path, res_type=ResolutionType.MODEL)

    # Check if the content hash is the same
    fingerprint = hash_model_results(results)
    if resolution.fingerprint != fingerprint:
        raise MatchboxResolutionInvalidData

    with MBDB.get_session() as session:
        existing_results = session.execute(
            select(func.count())
            .select_from(Results)
            .where(Results.resolution_id == resolution.resolution_id)
        ).scalar_one()

        if existing_results > 0:
            raise MatchboxResolutionExistingData

    logger.info(
        f"Writing results data with batch size {batch_size:,}", prefix=log_prefix
    )

    # Get a cluster lookup dictionary based on the resolution's parents
    im = IntMap()
    nested_data = get_parent_clusters_and_leaves(resolution=resolution)
    cluster_lookup: dict[int, Cluster] = _build_cluster_objects(nested_data, im)

    logger.debug("Computing hierarchies", prefix=log_prefix)
    cluster_df = _build_cluster_hierarchy(
        cluster_lookup=cluster_lookup, probabilities=results
    )
    del cluster_lookup

    logger.debug("Ingesting clusters dataframe", prefix=log_prefix)
    with (
        MBDB.get_session() as session,
        ingest_to_temporary_table(
            table_name="model_results",
            schema_name="mb",
            column_types={
                "left_id": BIGINT(),
                "right_id": BIGINT(),
                "probability": SMALLINT(),
            },
            data=results,
            max_chunksize=batch_size,
        ) as new_results,
        ingest_to_temporary_table(
            table_name="model_clusters",
            schema_name="mb",
            column_types={
                "cluster_id": BIGINT(),
                "cluster_hash": BYTEA(),
                "leaves": ARRAY(BIGINT),
                "probability": SMALLINT(),
                "is_new": BOOLEAN(),
            },
            data=cluster_df.with_columns(pl.lit(True).alias("is_new")).to_arrow(),
            max_chunksize=batch_size,
        ) as new_clusters,
    ):
        try:
            # Labelling new clusters is technically unnecessary but theory is that
            # in our case it will be faster than a lot of ignored conflicts
            session.execute(
                update(new_clusters).values(
                    is_new=~exists(
                        select(1).where(
                            Clusters.cluster_hash == new_clusters.c.cluster_hash
                        )
                    )
                )
            )
            session.flush()
            # Clusters
            cluster_select = (
                select(new_clusters.c.cluster_hash)
                .distinct()
                .where(new_clusters.c.is_new)
            )
            cluster_results = session.execute(
                insert(Clusters)
                .from_select(["cluster_hash"], cluster_select)
                .on_conflict_do_nothing(index_elements=[Clusters.cluster_hash])
            )
            logger.info(
                f"Will add {cluster_results.rowcount:,} entries to Clusters table",
                prefix=log_prefix,
            )
            session.flush()

            # Contains
            contains_select = (
                select(
                    Clusters.cluster_id.label("root"),
                    func.unnest(new_clusters.c.leaves).label("leaf"),
                )
                .select_from(
                    new_clusters.join(
                        Clusters, Clusters.cluster_hash == new_clusters.c.cluster_hash
                    )
                )
                .where(new_clusters.c.is_new)
            )
            contains_results = session.execute(
                insert(Contains)
                .from_select(["root", "leaf"], contains_select)
                .on_conflict_do_nothing(index_elements=[Contains.root, Contains.leaf])
            )
            logger.info(
                f"Will add {contains_results.rowcount:,} entries to Contains table",
                prefix=log_prefix,
            )

            # Probabilities
            prob_select = select(
                literal(resolution.resolution_id, BIGINT).label("resolution_id"),
                Clusters.cluster_id,
                new_clusters.c.probability,
            ).select_from(
                new_clusters.join(
                    Clusters, Clusters.cluster_hash == new_clusters.c.cluster_hash
                )
            )
            prob_res = session.execute(
                insert(Probabilities).from_select(
                    ["resolution_id", "cluster_id", "probability"], prob_select
                )
            )
            logger.info(
                f"Will add {prob_res.rowcount:,} entries to Probabilities table",
                prefix=log_prefix,
            )

            # Results
            results_select = select(
                literal(resolution.resolution_id, BIGINT).label("resolution_id"),
                new_results.c.left_id,
                new_results.c.right_id,
                new_results.c.probability,
            )
            results_res = session.execute(
                insert(Results).from_select(
                    ["resolution_id", "left_id", "right_id", "probability"],
                    results_select,
                )
            )
            logger.info(
                f"Will add {results_res.rowcount:,} entries to Results table",
                prefix=log_prefix,
            )
            session.commit()

        except Exception as e:
            logger.error(
                f"Failed to insert data, rolling back: {str(e)}", prefix=log_prefix
            )
            session.rollback()
            raise

    MBDB.vacuum_analyze(
        Clusters.__table__.fullname,
        Contains.__table__.fullname,
        Probabilities.__table__.fullname,
        Results.__table__.fullname,
    )

    logger.info("Insert operation complete!", prefix=log_prefix)
